from collections.abc import Sequence

import warnings

import dask.dataframe as dd

import dask
from dask.dataframe.dispatch import make_meta_dispatch
from dask.dataframe.backends import _nonempty_index, meta_nonempty, meta_nonempty_dataframe, _nonempty_series
from tape.utils import IndexCallable
from dask.dataframe.core import get_parallel_type
from dask.dataframe.extensions import make_array_nonempty

import numpy as np
import pandas as pd

from typing import Literal


from functools import partial
from dask.dataframe.io.parquet.arrow import (
    ArrowDatasetEngine as DaskArrowDatasetEngine,
)

SOURCE_FRAME_LABEL = "source"  # Reserved label for source table
OBJECT_FRAME_LABEL = "object"  # Reserved label for object table.

__all__ = [
    "EnsembleFrame",
    "EnsembleSeries",
    "ObjectFrame",
    "SourceFrame",
    "TapeFrame",
    "TapeObjectFrame",
    "TapeSourceFrame",
    "TapeSeries",
]


class TapeArrowEngine(DaskArrowDatasetEngine):
    """
    Engine for reading parquet files into Tape and assigning the appropriate Dask meta.

    Based off of the approach used in dask_geopandas.io
    """

    @classmethod
    def _creates_meta(cls, meta, schema):
        """
        Converts the meta to a TapeFrame.
        """
        return TapeFrame(meta)

    @classmethod
    def _create_dd_meta(cls, dataset_info, use_nullable_dtypes=False):
        """Overriding private method for dask >= 2021.10.0"""
        meta = super()._create_dd_meta(dataset_info)

        schema = dataset_info["schema"]
        if not schema.names and not schema.metadata:
            if len(list(dataset_info["ds"].get_fragments())) == 0:
                raise ValueError(
                    "No dataset parts discovered. Use dask.dataframe.read_parquet "
                    "to read it as an empty DataFrame"
                )
        meta = cls._creates_meta(meta, schema)
        return meta


class TapeSourceArrowEngine(TapeArrowEngine):
    """
    Barebones subclass of TapeArrowEngine for assigning the meta when loading from a parquet file
    of source data.
    """

    @classmethod
    def _creates_meta(cls, meta, schema):
        """
        Convert meta to a TapeSourceFrame
        """
        return TapeSourceFrame(meta)


class TapeObjectArrowEngine(TapeArrowEngine):
    """
    Barebones subclass of TapeArrowEngine for assigning the meta when loading from a parquet file
    of object data.
    """

    @classmethod
    def _creates_meta(cls, meta, schema):
        """
        Convert meta to a TapeObjectFrame
        """
        return TapeObjectFrame(meta)


class _Frame(dd.core._Frame):
    """Base class for extensions of Dask Dataframes that track additional Ensemble-related metadata."""

    def __init__(self, dsk, name, meta, divisions, label=None, ensemble=None):
        # We define relevant object fields before super().__init__ since that call may lead to a
        # map_partitions call which will assume these fields exist.
        self.label = label  # A label used by the Ensemble to identify this frame.
        self.ensemble = ensemble  # The Ensemble object containing this frame.
        self.dirty = False  # True if the underlying data is out of sync with the Ensemble

        super().__init__(dsk, name, meta, divisions)

    def is_dirty(self):
        return self.dirty

    def set_dirty(self, dirty):
        self.dirty = dirty

    @property
    def _args(self):
        # Ensure our Dask extension can correctly be used by pickle.
        # See https://github.com/geopandas/dask-geopandas/issues/237
        return super()._args + (self.label, self.ensemble)

    @property
    def partitions(self):
        """Slice dataframe by partitions

        This allows partitionwise slicing of a TAPE EnsembleFrame.  You can perform normal
        Numpy-style slicing, but now rather than slice elements of the array you
        slice along partitions so, for example, ``df.partitions[:5]`` produces a new
        Dask Dataframe of the first five partitions. Valid indexers are integers, sequences
        of integers, slices, or boolean masks.

        Examples
        --------
        >>> df.partitions[0]  # doctest: +SKIP
        >>> df.partitions[:3]  # doctest: +SKIP
        >>> df.partitions[::10]  # doctest: +SKIP

        Returns
        -------
        A TAPE EnsembleFrame Object
        """
        self.set_dirty(True)
        return IndexCallable(self._partitions, self.is_dirty(), self.ensemble)

    def _propagate_metadata(self, new_frame):
        """Propagates any relevant metadata to a new frame.

        Parameters
        ----------
        new_frame: `_Frame`
            A frame to propage metadata to

        Returns
        ----------
        new_frame: `_Frame`
            The modifed frame
        """
        new_frame.label = self.label
        new_frame.ensemble = self.ensemble
        new_frame.set_dirty(self.is_dirty())
        return new_frame

    def copy(self):
        self_copy = super().copy()
        return self._propagate_metadata(self_copy)

    def assign(self, **kwargs):
        """Assign new columns to a DataFrame.

        This docstring was copied from dask.dataframe.DataFrame.assign.

        Some inconsistencies with the Dask version may exist.

        Returns a new object with all original columns in addition to new ones. Existing columns
        that are re-assigned will be overwritten.

        Parameters
        ----------
        **kwargs: `dict`
            The column names are keywords. If the values are callable, they are computed on the
            DataFrame and assigned to the new columns. The callable must not change input DataFrame
            (though pandas doesn't check it). If the values are not callable, (e.g. a Series,
            scalar, or array), they are simply assigned.

        Returns
        ----------
        result: `tape._Frame`
            The modifed frame
        """
        result = self._propagate_metadata(super().assign(**kwargs))
        result.set_dirty(True)
        return result

    def query(self, expr, **kwargs):
        """Filter dataframe with complex expression

        Doc string below derived from dask.dataframe.core

        Blocked version of pd.DataFrame.query

        Parameters
        ----------
        expr: str
            The query string to evaluate.
            You can refer to column names that are not valid Python variable names
            by surrounding them in backticks.
            Dask does not fully support referring to variables using the '@' character,
            use f-strings or the ``local_dict`` keyword argument instead.
        **kwargs: `dict`
            See the documentation for eval() for complete details on the keyword arguments accepted
            by pandas.DataFrame.query().

        Returns
        ----------
        result: `tape._Frame`
            The modifed frame

        Notes
        -----
        This is like the sequential version except that this will also happen
        in many threads.  This may conflict with ``numexpr`` which will use
        multiple threads itself.  We recommend that you set ``numexpr`` to use a
        single thread:

        .. code-block:: python

            import numexpr
            numexpr.set_num_threads(1)
        """
        result = self._propagate_metadata(super().query(expr, **kwargs))
        result.set_dirty(True)
        return result

    def sample(self, **kwargs):
        """Random sample of items from a Dataframe.

        Doc string below derived from dask.dataframe.core

        Parameters
        ----------
        frac: float, optional
            Approximate fraction of objects to return. This sampling fraction
            is applied to all partitions equally. Note that this is an
            approximate fraction. You should not expect exactly len(df) * frac
            items to be returned, as the exact number of elements selected will
            depend on how your data is partitioned (but should be pretty close
            in practice).
        replace: boolean, optional
            Sample with or without replacement. Default = False.
        random_state: int or np.random.RandomState
            If an int, we create a new RandomState with this as the seed;
            Otherwise we draw from the passed RandomState.

        Returns
        ----------
        result: `tape._Frame`
            The modifed frame

        """
        result = self._propagate_metadata(super().sample(**kwargs))
        result.set_dirty(True)
        return result

    def merge(self, right, **kwargs):
        """Merge the Dataframe with another DataFrame

        Doc string below derived from dask.dataframe.core

        This will merge the two datasets, either on the indices, a certain column
        in each dataset or the index in one dataset and the column in another.

        Parameters
        ----------
        right: dask.dataframe.DataFrame
        how : {'left', 'right', 'outer', 'inner'}, default: 'inner'
            How to handle the operation of the two objects:

            - left: use calling frame's index (or column if on is specified)
            - right: use other frame's index
            - outer: form union of calling frame's index (or column if on is
              specified) with other frame's index, and sort it
              lexicographically
            - inner: form intersection of calling frame's index (or column if
              on is specified) with other frame's index, preserving the order
              of the calling's one

        on : label or list
            Column or index level names to join on. These must be found in both
            DataFrames. If on is None and not merging on indexes then this
            defaults to the intersection of the columns in both DataFrames.
        left_on : label or list, or array-like
            Column to join on in the left DataFrame. Other than in pandas
            arrays and lists are only support if their length is 1.
        right_on : label or list, or array-like
            Column to join on in the right DataFrame. Other than in pandas
            arrays and lists are only support if their length is 1.
        left_index : boolean, default False
            Use the index from the left DataFrame as the join key.
        right_index : boolean, default False
            Use the index from the right DataFrame as the join key.
        suffixes : 2-length sequence (tuple, list, ...)
            Suffix to apply to overlapping column names in the left and
            right side, respectively
        indicator : boolean or string, default False
            If True, adds a column to output DataFrame called "_merge" with
            information on the source of each row. If string, column with
            information on source of each row will be added to output DataFrame,
            and column will be named value of string. Information column is
            Categorical-type and takes on a value of "left_only" for observations
            whose merge key only appears in `left` DataFrame, "right_only" for
            observations whose merge key only appears in `right` DataFrame,
            and "both" if the observation's merge key is found in both.
        npartitions: int or None, optional
            The ideal number of output partitions. This is only utilised when
            performing a hash_join (merging on columns only). If ``None`` then
            ``npartitions = max(lhs.npartitions, rhs.npartitions)``.
            Default is ``None``.
        shuffle: {'disk', 'tasks', 'p2p'}, optional
            Either ``'disk'`` for single-node operation or ``'tasks'`` and
            ``'p2p'``` for distributed operation.  Will be inferred by your
            current scheduler.
        broadcast: boolean or float, optional
            Whether to use a broadcast-based join in lieu of a shuffle-based
            join for supported cases.  By default, a simple heuristic will be
            used to select the underlying algorithm. If a floating-point value
            is specified, that number will be used as the ``broadcast_bias``
            within the simple heuristic (a large number makes Dask more likely
            to choose the ``broacast_join`` code path). See ``broadcast_join``
            for more information.

        Notes
        -----

        There are three ways to join dataframes:

        1. Joining on indices. In this case the divisions are
           aligned using the function ``dask.dataframe.multi.align_partitions``.
           Afterwards, each partition is merged with the pandas merge function.

        2. Joining one on index and one on column. In this case the divisions of
           dataframe merged by index (:math:`d_i`) are used to divide the column
           merged dataframe (:math:`d_c`) one using
           ``dask.dataframe.multi.rearrange_by_divisions``. In this case the
           merged dataframe (:math:`d_m`) has the exact same divisions
           as (:math:`d_i`). This can lead to issues if you merge multiple rows from
           (:math:`d_c`) to one row in (:math:`d_i`).

        3. Joining both on columns. In this case a hash join is performed using
           ``dask.dataframe.multi.hash_join``.

        In some cases, you may see a ``MemoryError`` if the ``merge`` operation requires
        an internal ``shuffle``, because shuffling places all rows that have the same
        index in the same partition. To avoid this error, make sure all rows with the
        same ``on``-column value can fit on a single partition.
        """
        result = super().merge(right, **kwargs)
        return self._propagate_metadata(result)

    def join(self, other, **kwargs):
        """Join columns of another DataFrame. Note that if `other` is a different type,
        we expect the result to have the type of this object regardless of the value
        of the`how` parameter.

        This docstring was copied from pandas.core.frame.DataFrame.join.

        Some inconsistencies with this version may exist.

        Join columns with `other` DataFrame either on index or on a key
        column. Efficiently join multiple DataFrame objects by index at once by
        passing a list.

        Parameters
        ----------
        other : DataFrame, Series, or a list containing any combination of them
            Index should be similar to one of the columns in this one. If a
            Series is passed, its name attribute must be set, and that will be
            used as the column name in the resulting joined DataFrame.
        on : str, list of str, or array-like, optional
            Column or index level name(s) in the caller to join on the index
            in `other`, otherwise joins index-on-index. If multiple
            values given, the `other` DataFrame must have a MultiIndex. Can
            pass an array as the join key if it is not already contained in
            the calling DataFrame. Like an Excel VLOOKUP operation.
        how : {'left', 'right', 'outer', 'inner', 'cross'}, default 'left'
            How to handle the operation of the two objects.

            * left: use calling frame's index (or column if on is specified)
            * right: use `other`'s index.
            * outer: form union of calling frame's index (or column if on is
              specified) with `other`'s index, and sort it lexicographically.
            * inner: form intersection of calling frame's index (or column if
              on is specified) with `other`'s index, preserving the order
              of the calling's one.
            * cross: creates the cartesian product from both frames, preserves the order
              of the left keys.
        lsuffix : str, default ''
            Suffix to use from left frame's overlapping columns.
        rsuffix : str, default ''
            Suffix to use from right frame's overlapping columns.
        sort : bool, default False
            Order result DataFrame lexicographically by the join key. If False,
            the order of the join key depends on the join type (how keyword).
        validate : str, optional
            If specified, checks if join is of specified type.

            * "one_to_one" or "1:1": check if join keys are unique in both left
              and right datasets.
            * "one_to_many" or "1:m": check if join keys are unique in left dataset.
            * "many_to_one" or "m:1": check if join keys are unique in right dataset.
            * "many_to_many" or "m:m": allowed, but does not result in checks.

        Returns
        -------
        result: `tape._Frame`
            A TAPE dataframe containing columns from both the caller and `other`.

        """
        result = super().join(other, **kwargs)
        return self._propagate_metadata(result)

    def drop(self, labels=None, axis=0, columns=None, errors="raise"):
        """Drop specified labels from rows or columns.

        Doc string below derived from dask.dataframe.core

        Remove rows or columns by specifying label names and corresponding
        axis, or by directly specifying index or column names. When using a
        multi-index, labels on different levels can be removed by specifying
        the level. See the :ref:`user guide <advanced.shown_levels>`
        for more information about the now unused levels.

        Parameters
        ----------
        labels : single label or list-like
            Index or column labels to drop. A tuple will be used as a single
            label and not treated as a list-like.
        axis : {0 or 'index', 1 or 'columns'}, default 0
            Whether to drop labels from the index (0 or 'index') or
            columns (1 or 'columns').
            is equivalent to ``index=labels``.
        columns : single label or list-like
            Alternative to specifying axis (``labels, axis=1``
            is equivalent to ``columns=labels``).
        errors : {'ignore', 'raise'}, default 'raise'
            If 'ignore', suppress error and only existing labels are
            dropped.

        Returns
        -------
        result: `tape._Frame`
            Returns the frame or None with the specified
            index or column labels removed or None if inplace=True.
        """
        result = self._propagate_metadata(
            super().drop(labels=labels, axis=axis, columns=columns, errors=errors)
        )
        result.set_dirty(True)
        return result

    def dropna(self, **kwargs):
        """
        Remove missing values.

        Doc string below derived from dask.dataframe.core

        Parameters
        ----------

        how : {'any', 'all'}, default 'any'
            Determine if row or column is removed from DataFrame, when we have
            at least one NA or all NA.

            * 'any' : If any NA values are present, drop that row or column.
            * 'all' : If all values are NA, drop that row or column.

        thresh : int, optional
            Require that many non-NA values. Cannot be combined with how.
        subset : column label or sequence of labels, optional
            Labels along other axis to consider, e.g. if you are dropping rows
            these would be a list of columns to include.

        Returns
        ----------
        result: `tape._Frame`
            The modifed frame with NA entries dropped from it or None if ``inplace=True``.
        """
        result = self._propagate_metadata(super().dropna(**kwargs))
        result.set_dirty(True)
        return result

    def persist(self, **kwargs):
        """Persist this dask collection into memory

        Doc string below derived from dask.base

        This turns a lazy Dask collection into a Dask collection with the same
        metadata, but now with the results fully computed or actively computing
        in the background.

        The action of function differs significantly depending on the active
        task scheduler.  If the task scheduler supports asynchronous computing,
        such as is the case of the dask.distributed scheduler, then persist
        will return *immediately* and the return value's task graph will
        contain Dask Future objects.  However if the task scheduler only
        supports blocking computation then the call to persist will *block*
        and the return value's task graph will contain concrete Python results.

        This function is particularly useful when using distributed systems,
        because the results will be kept in distributed memory, rather than
        returned to the local process as with compute.

        Parameters
        ----------
        **kwargs
            Extra keywords to forward to the scheduler function.

        Returns
        -------
        result: `tape._Frame`
            The modifed frame backed by in-memory data
        """
        result = super().persist(**kwargs)
        return self._propagate_metadata(result)

    def set_index(
        self,
        other,
        drop=True,
        sorted=False,
        npartitions=None,
        divisions=None,
        inplace=False,
        sort=True,
        **kwargs,
    ):
        """Set the DataFrame index (row labels) using an existing column.

        Doc string below derived from dask.dataframe.core

        If ``sort=False``, this function operates exactly like ``pandas.set_index``
        and sets the index on the DataFrame. If ``sort=True`` (default),
        this function also sorts the DataFrame by the new index. This can have a
        significant impact on performance, because joins, groupbys, lookups, etc.
        are all much faster on that column. However, this performance increase
        comes with a cost, sorting a parallel dataset requires expensive shuffles.
        Often we ``set_index`` once directly after data ingest and filtering and
        then perform many cheap computations off of the sorted dataset.

        With ``sort=True``, this function is much more expensive. Under normal
        operation this function does an initial pass over the index column to
        compute approximate quantiles to serve as future divisions. It then passes
        over the data a second time, splitting up each input partition into several
        pieces and sharing those pieces to all of the output partitions now in
        sorted order.

        In some cases we can alleviate those costs, for example if your dataset is
        sorted already then we can avoid making many small pieces or if you know
        good values to split the new index column then we can avoid the initial
        pass over the data. For example if your new index is a datetime index and
        your data is already sorted by day then this entire operation can be done
        for free. You can control these options with the following parameters.

        Parameters
        ----------
        other: string or Dask Series
            Column to use as index.
        drop: boolean, default True
            Delete column to be used as the new index.
        sorted: bool, optional
            If the index column is already sorted in increasing order.
            Defaults to False
        npartitions: int, None, or 'auto'
            The ideal number of output partitions. If None, use the same as
            the input. If 'auto' then decide by memory use.
            Only used when ``divisions`` is not given. If ``divisions`` is given,
            the number of output partitions will be ``len(divisions) - 1``.
        divisions: list, optional
            The "dividing lines" used to split the new index into partitions.
            For ``divisions=[0, 10, 50, 100]``, there would be three output partitions,
            where the new index contained [0, 10), [10, 50), and [50, 100), respectively.
            See https://docs.dask.org/en/latest/dataframe-design.html#partitions.
            If not given (default), good divisions are calculated by immediately computing
            the data and looking at the distribution of its values. For large datasets,
            this can be expensive.
            Note that if ``sorted=True``, specified divisions are assumed to match
            the existing partitions in the data; if this is untrue you should
            leave divisions empty and call ``repartition`` after ``set_index``.
        inplace: bool, optional
            Modifying the DataFrame in place is not supported by Dask.
            Defaults to False.
        sort: bool, optional
            If ``True``, sort the DataFrame by the new index. Otherwise
            set the index on the individual existing partitions.
            Defaults to ``True``.
        shuffle: {'disk', 'tasks', 'p2p'}, optional
            Either ``'disk'`` for single-node operation or ``'tasks'`` and
            ``'p2p'`` for distributed operation.  Will be inferred by your
            current scheduler.
        compute: bool, default False
            Whether or not to trigger an immediate computation. Defaults to False.
            Note, that even if you set ``compute=False``, an immediate computation
            will still be triggered if ``divisions`` is ``None``.
        partition_size: int, optional
            Desired size of each partitions in bytes.
            Only used when ``npartitions='auto'``

        Returns
        ----------
        result: `tape._Frame`
            The indexed frame
        """
        result = super().set_index(other, drop, sorted, npartitions, divisions, inplace, sort, **kwargs)
        return self._propagate_metadata(result)

    def map_partitions(self, func, *args, **kwargs):
        """Apply Python function on each DataFrame partition.

        Doc string below derived from dask.dataframe.core

        If ``sort=False``, this function operates exactly like ``pandas.set_index``
        and sets the index on the DataFrame. If ``sort=True`` (default),
        this function also sorts the DataFrame by the new index. This can have a
        significant impact on performance, because joins, groupbys, lookups, etc.
        are all much faster on that column. However, this performance increase
        comes with a cost, sorting a parallel dataset requires expensive shuffles.
        Often we ``set_index`` once directly after data ingest and filtering and
        then perform many cheap computations off of the sorted dataset.

        With ``sort=True``, this function is much more expensive. Under normal
        operation this function does an initial pass over the index column to
        compute approximate quantiles to serve as future divisions. It then passes
        over the data a second time, splitting up each input partition into several
        pieces and sharing those pieces to all of the output partitions now in
        sorted order.

        In some cases we can alleviate those costs, for example if your dataset is
        sorted already then we can avoid making many small pieces or if you know
        good values to split the new index column then we can avoid the initial
        pass over the data. For example if your new index is a datetime index and
        your data is already sorted by day then this entire operation can be done
        for free. You can control these options with the following parameters.

        Parameters
        ----------
        other: string or Dask Series
            Column to use as index.
        drop: boolean, default True
            Delete column to be used as the new index.
        sorted: bool, optional
            If the index column is already sorted in increasing order.
            Defaults to False
        npartitions: int, None, or 'auto'
            The ideal number of output partitions. If None, use the same as
            the input. If 'auto' then decide by memory use.
            Only used when ``divisions`` is not given. If ``divisions`` is given,
            the number of output partitions will be ``len(divisions) - 1``.
        divisions: list, optional
            The "dividing lines" used to split the new index into partitions.
            For ``divisions=[0, 10, 50, 100]``, there would be three output partitions,
            where the new index contained [0, 10), [10, 50), and [50, 100), respectively.
            See https://docs.dask.org/en/latest/dataframe-design.html#partitions.
            If not given (default), good divisions are calculated by immediately computing
            the data and looking at the distribution of its values. For large datasets,
            this can be expensive.
            Note that if ``sorted=True``, specified divisions are assumed to match
            the existing partitions in the data; if this is untrue you should
            leave divisions empty and call ``repartition`` after ``set_index``.
        inplace: bool, optional
            Modifying the DataFrame in place is not supported by Dask.
            Defaults to False.
        sort: bool, optional
            If ``True``, sort the DataFrame by the new index. Otherwise
            set the index on the individual existing partitions.
            Defaults to ``True``.
        shuffle: {'disk', 'tasks', 'p2p'}, optional
            Either ``'disk'`` for single-node operation or ``'tasks'`` and
            ``'p2p'`` for distributed operation.  Will be inferred by your
            current scheduler.
        compute: bool, default False
            Whether or not to trigger an immediate computation. Defaults to False.
            Note, that even if you set ``compute=False``, an immediate computation
            will still be triggered if ``divisions`` is ``None``.
        partition_size: int, optional
            Desired size of each partitions in bytes.
            Only used when ``npartitions='auto'``
        """
        result = super().map_partitions(func, *args, **kwargs)
        if isinstance(result, self.__class__):
            # If the output of func is another _Frame, let's propagate any metadata.
            return self._propagate_metadata(result)
        return result

    def compute(self, **kwargs):
        """Compute this Dask collection, returning the underlying dataframe or series.
        If tracked by an `Ensemble`, the `Ensemble` is informed of this operation and
        is given the opportunity to sync any of its tables prior to this Dask collection
        being computed.

        Doc string below derived from dask.dataframe.DataFrame.compute

        This turns a lazy Dask collection into its in-memory equivalent. For example
        a Dask array turns into a NumPy array and a Dask dataframe turns into a
        Pandas dataframe. The entire dataset must fit into memory before calling
        this operation.

        Parameters
        ----------
        scheduler: `string`, optional
            Which scheduler to use like “threads”, “synchronous” or “processes”.
            If not provided, the default is to check the global settings first,
            and then fall back to the collection defaults.
        optimize_graph: `bool`, optional
            If True [default], the graph is optimized before computation.
            Otherwise the graph is run as is. This can be useful for debugging.
        **kwargs: `dict`, optional
            Extra keywords to forward to the scheduler function.
        """
        if self.ensemble is not None:
            self.ensemble._lazy_sync_tables_from_frame(self)
        return super().compute(**kwargs)

    def repartition(
        self,
        divisions=None,
        npartitions=None,
        partition_size=None,
        freq=None,
        force=False,
    ):
        """Repartition dataframe along new divisions

        Doc string below derived from dask.dataframe.DataFrame

        Parameters
        ----------
        divisions : list, optional
            The "dividing lines" used to split the dataframe into partitions.
            For ``divisions=[0, 10, 50, 100]``, there would be three output partitions,
            where the new index contained [0, 10), [10, 50), and [50, 100), respectively.
            See https://docs.dask.org/en/latest/dataframe-design.html#partitions.
            Only used if npartitions and partition_size isn't specified.
            For convenience if given an integer this will defer to npartitions
            and if given a string it will defer to partition_size (see below)
        npartitions : int, optional
            Approximate number of partitions of output. Only used if partition_size
            isn't specified. The number of partitions used may be slightly
            lower than npartitions depending on data distribution, but will never be
            higher.
        partition_size: int or string, optional
            Max number of bytes of memory for each partition. Use numbers or
            strings like 5MB. If specified npartitions and divisions will be
            ignored. Note that the size reflects the number of bytes used as
            computed by ``pandas.DataFrame.memory_usage``, which will not
            necessarily match the size when storing to disk.

            .. warning::

               This keyword argument triggers computation to determine
               the memory size of each partition, which may be expensive.

        freq : str, pd.Timedelta
            A period on which to partition timeseries data like ``'7D'`` or
            ``'12h'`` or ``pd.Timedelta(hours=12)``.  Assumes a datetime index.
        force : bool, default False
            Allows the expansion of the existing divisions.
            If False then the new divisions' lower and upper bounds must be
            the same as the old divisions'.

        Notes
        -----
        Exactly one of `divisions`, `npartitions`, `partition_size`, or `freq`
        should be specified. A ``ValueError`` will be raised when that is
        not the case.

        Also note that ``len(divisons)`` is equal to ``npartitions + 1``. This is because ``divisions``
        represents the upper and lower bounds of each partition. The first item is the
        lower bound of the first partition, the second item is the lower bound of the
        second partition and the upper bound of the first partition, and so on.
        The second-to-last item is the lower bound of the last partition, and the last
        (extra) item is the upper bound of the last partition.

        Examples
        --------
        >>> df = df.repartition(npartitions=10)  # doctest: +SKIP
        >>> df = df.repartition(divisions=[0, 5, 10, 20])  # doctest: +SKIP
        >>> df = df.repartition(freq='7d')  # doctest: +SKIP
        """
        result = super().repartition(
            divisions=divisions,
            npartitions=npartitions,
            partition_size=partition_size,
            freq=freq,
            force=force,
        )
        return self._propagate_metadata(result)


class TapeSeries(pd.Series):
    """A barebones extension of a Pandas series to be used for underlying Ensemble data.

    See https://pandas.pydata.org/docs/development/extending.html#subclassing-pandas-data-structures
    """

    @property
    def _constructor(self):
        return TapeSeries

    @property
    def _constructor_sliced(self):
        return TapeSeries


class TapeFrame(pd.DataFrame):
    """A barebones extension of a Pandas frame to be used for underlying Ensemble data.

    See https://pandas.pydata.org/docs/development/extending.html#subclassing-pandas-data-structures
    """

    @property
    def _constructor(self):
        return TapeFrame

    @property
    def _constructor_expanddim(self):
        return TapeFrame


class EnsembleSeries(_Frame, dd.core.Series):
    """A barebones extension of a Dask Series for Ensemble data."""

    _partition_type = TapeSeries  # Tracks the underlying data type


class EnsembleFrame(_Frame, dd.core.DataFrame):
    """An extension for a Dask Dataframe for data used by a lightcurve Ensemble.

    The underlying non-parallel dataframes are TapeFrames and TapeSeries which extend Pandas frames.

    Examples
    ----------
    Instatiation::

        import tape
        ens = tape.Ensemble()
        data = {...} # Some data you want tracked by the Ensemble
        ensemble_frame = tape.EnsembleFrame.from_dict(data, label="my_frame", ensemble=ens)
    """

    _partition_type = TapeFrame  # Tracks the underlying data type

    def __getitem__(self, key):
        result = super().__getitem__(key)
        if isinstance(result, _Frame):
            # Ensures that any _Frame metadata is propagated.
            result = self._propagate_metadata(result)
        return result

    @classmethod
    def from_tapeframe(cls, data, npartitions=None, chunksize=None, sort=True, label=None, ensemble=None):
        """Returns an EnsembleFrame constructed from a TapeFrame.

        Parameters
        ----------
        data: `TapeFrame`
            Frame containing the underlying data fro the EnsembleFram
        npartitions: `int`, optional
            The number of partitions of the index to create. Note that depending on
            the size and index of the dataframe, the output may have fewer
            partitions than requested.
        chunksize: `int`, optional
            Size of the individual chunks of data in non-parallel objects that make up Dask frames.
        sort: `bool`, optional
            Whether to sort the frame by a default index.
        label: `str`, optional
            The label used to by the Ensemble to identify the frame.
        ensemble: `tape.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.EnsembleFrame`
            The constructed EnsembleFrame object.
        """
        result = dd.from_pandas(data, npartitions=npartitions, chunksize=chunksize, sort=sort)
        result.label = label
        result.ensemble = ensemble
        return result

    @classmethod
    def from_dask_dataframe(cl, df, ensemble=None, label=None):
        """Returns an EnsembleFrame constructed from a Dask dataframe.

        Parameters
        ----------
        df: `dask.dataframe.DataFrame` or `list`
            a Dask dataframe to convert to an EnsembleFrame
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.
        label: `str`, optional
            The label used to by the Ensemble to identify the frame.

        Returns
        ----------
        result: `tape.EnsembleFrame`
            The constructed EnsembleFrame object.
        """
        # Create a EnsembleFrame by mapping the partitions to the appropriate meta, TapeFrame
        # TODO(wbeebe@uw.edu): Determine if there is a better method
        result = df.map_partitions(TapeFrame)
        result.ensemble = ensemble
        result.label = label
        return result

    def update_ensemble(self):
        """Updates the Ensemble linked by the `EnsembelFrame.ensemble` property to track this frame.

        Returns
        ----------
        result: `tape.Ensemble`
            The Ensemble object which tracks this frame, `None` if no such Ensemble.
        """
        if self.ensemble is None:
            return None
        # Update the Ensemble to track this frame and return the ensemble.
        return self.ensemble.update_frame(self)

    def convert_flux_to_mag(
        self,
        flux_col,
        zero_point,
        err_col=None,
        zp_form="mag",
        out_col_name=None,
    ):
        """Converts this EnsembleFrame's flux column into a magnitude column, returning a new
        EnsembleFrame.

        Parameters
        ----------
        flux_col: 'str'
            The name of the EnsembleFrame flux column to convert into magnitudes.
        zero_point: 'str'
            The name of the EnsembleFrame column containing the zero point
            information for column transformation.
        err_col: 'str', optional
            The name of the EnsembleFrame column containing the errors to propagate.
            Errors are propagated using the following approximation:
            Err= (2.5/log(10))*(flux_error/flux), which holds mainly when the
            error in flux is much smaller than the flux.
        zp_form: `str`, optional
            The form of the zero point column, either "flux" or
            "magnitude"/"mag". Determines how the zero point (zp) is applied in
            the conversion. If "flux", then the function is applied as
            mag=-2.5*log10(flux/zp), or if "magnitude", then
            mag=-2.5*log10(flux)+zp.
        out_col_name: 'str', optional
            The name of the output magnitude column, if None then the output
            is just the flux column name + "_mag". The error column is also
            generated as the out_col_name + "_err".

        Returns
        ----------
        result: `tape.EnsembleFrame`
            A new EnsembleFrame object with a new magnitude (and error) column.
        """
        if out_col_name is None:
            out_col_name = flux_col + "_mag"

        result = None
        if zp_form == "flux":  # mag = -2.5*np.log10(flux/zp)
            result = self.assign(**{out_col_name: lambda x: -2.5 * np.log10(x[flux_col] / x[zero_point])})

        elif zp_form == "magnitude" or zp_form == "mag":  # mag = -2.5*np.log10(flux) + zp
            result = self.assign(**{out_col_name: lambda x: -2.5 * np.log10(x[flux_col]) + x[zero_point]})
        else:
            raise ValueError(f"{zp_form} is not a valid zero_point format.")

        # Calculate Errors
        if err_col is not None:
            result = result.assign(
                **{out_col_name + "_err": lambda x: (2.5 / np.log(10)) * (x[err_col] / x[flux_col])}
            )

        return result

    def coalesce(self, input_cols, output_col, drop_inputs=False):
        """Combines multiple input columns into a single output column, with
        values equal to the first non-nan value encountered in the input cols.

        Parameters
        ----------
        input_cols: `list`
            The list of column names to coalesce into a single column.
        output_col: `str`, optional
            The name of the coalesced output column.
        drop_inputs: `bool`, optional
            Determines whether the input columns are dropped or preserved. If
            a mapped column is an input and dropped, the output column is
            automatically assigned to replace that column mapping internally.

        Returns
        -------
        ensemble: `tape.ensemble.Ensemble`
            An ensemble object.
        """

        def coalesce_partition(df, input_cols, output_col):
            """Coalescing function for a single partition (pandas dataframe)"""

            # Create a subset dataframe per input column
            # Rename column to output to allow combination
            input_dfs = []
            for col in input_cols:
                col_df = df[[col]]
                input_dfs.append(col_df.rename(columns={col: output_col}))

            # Combine each dataframe
            coal_df = input_dfs.pop()
            while input_dfs:
                coal_df = coal_df.combine_first(input_dfs.pop())

            # Assign the output column to the partition dataframe
            out_df = df.assign(**{output_col: coal_df[output_col]})

            return out_df

        table_ddf = self.map_partitions(lambda x: coalesce_partition(x, input_cols, output_col))

        # Drop the input columns if wanted
        if drop_inputs:
            if self.ensemble is not None:
                # First check to see if any dropped columns were critical columns
                current_map = self.ensemble.make_column_map().map
                cols_to_update = [key for key in current_map if current_map[key] in input_cols]

                # Theoretically a user could assign multiple critical columns in the input cols, this is very
                # likely to be a mistake, so we throw a warning here to alert them.
                if len(cols_to_update) > 1:
                    warnings.warn(
                        """Warning: Coalesce (with column dropping) is needing to update more than one
                    critical column mapping, please check that the resulting mapping is set as intended"""
                    )

                # Update critical columns to the new output column as needed
                if len(cols_to_update):  # if not zero
                    new_map = current_map
                    for col in cols_to_update:
                        new_map[col] = output_col

                    new_colmap = self.ensemble.make_column_map()
                    new_colmap.map = new_map

                    # Update the mapping
                    self.ensemble.update_column_mapping(new_colmap)

            table_ddf = table_ddf.drop(columns=input_cols)

        return table_ddf

    @classmethod
    def from_parquet(cl, path, index=None, columns=None, label=None, ensemble=None, **kwargs):
        """Returns an EnsembleFrame constructed from loading a parquet file.

        Parameters
        ----------
        path: `str` or `list`
            Source directory for data, or path(s) to individual parquet files. Prefix with a
            protocol like s3:// to read from alternative filesystems. To read from multiple
            files you can pass a globstring or a list of paths, with the caveat that they must all
            have the same protocol.
        index: `str`, `list`, `False`, optional
            Field name(s) to use as the output frame index. Default is None and index will be
            inferred from the pandas parquet file metadata, if present. Use False to read all
            fields as columns.
        columns: `str` or `list`, optional
            Field name(s) to read in as columns in the output. By default all non-index fields will
            be read (as determined by the pandas parquet metadata, if present). Provide a single
            field name instead of a list to read in the data as a Series.
        label: `str`, optional
            The label used to by the Ensemble to identify the frame.
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.EnsembleFrame`
            The constructed EnsembleFrame object.
        """
        # Read the parquet file with an engine that will assume the meta is a TapeFrame which Dask will
        # instantiate as EnsembleFrame via its dispatcher.
        result = dd.read_parquet(
            path, index=index, columns=columns, split_row_groups=True, engine=TapeArrowEngine, **kwargs
        )
        result.label = label
        result.ensemble = ensemble

        return result


class TapeSourceFrame(TapeFrame):
    """A barebones extension of a Pandas frame to be used for underlying Ensemble source data

    See https://pandas.pydata.org/docs/development/extending.html#subclassing-pandas-data-structures
    """

    @property
    def _constructor(self):
        return TapeSourceFrame

    @property
    def _constructor_expanddim(self):
        return TapeSourceFrame


class TapeObjectFrame(TapeFrame):
    """A barebones extension of a Pandas frame to be used for underlying Ensemble object data.

    See https://pandas.pydata.org/docs/development/extending.html#subclassing-pandas-data-structures
    """

    @property
    def _constructor(self):
        return TapeObjectFrame

    @property
    def _constructor_expanddim(self):
        return TapeObjectFrame


class SourceFrame(EnsembleFrame):
    """A subclass of EnsembleFrame for Source data."""

    _partition_type = TapeSourceFrame  # Tracks the underlying data type

    def __init__(self, dsk, name, meta, divisions, ensemble=None):
        super().__init__(dsk, name, meta, divisions)
        self.label = SOURCE_FRAME_LABEL  # A label used by the Ensemble to identify this frame.
        self.ensemble = ensemble  # The Ensemble object containing this frame.

    def __getitem__(self, key):
        result = super().__getitem__(key)
        if isinstance(result, _Frame):
            # Ensures that we have any metadata
            result = self._propagate_metadata(result)
        return result

    @classmethod
    def from_parquet(
        cl,
        path,
        index=None,
        columns=None,
        ensemble=None,
    ):
        """Returns a SourceFrame constructed from loading a parquet file.

        Parameters
        ----------
        path: `str` or `list`
            Source directory for data, or path(s) to individual parquet files. Prefix with a
            protocol like s3:// to read from alternative filesystems. To read from multiple
            files you can pass a globstring or a list of paths, with the caveat that they must all
            have the same protocol.
        columns: `str` or `list`, optional
            Field name(s) to read in as columns in the output. By default all non-index fields will
            be read (as determined by the pandas parquet metadata, if present). Provide a single
            field name instead of a list to read in the data as a Series.
        index: `str`, `list`, `False`, optional
            Field name(s) to use as the output frame index. Default is None and index will be
            inferred from the pandas parquet file metadata, if present. Use False to read all
            fields as columns.
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.EnsembleFrame`
            The constructed EnsembleFrame object.
        """
        # Read the source parquet file with an engine that will assume the meta is a
        # TapeSourceFrame which tells Dask to instantiate a SourceFrame via its
        # dispatcher.
        result = dd.read_parquet(
            path,
            index=index,
            columns=columns,
            split_row_groups=True,
            engine=TapeSourceArrowEngine,
        )
        result.ensemble = ensemble
        result.label = SOURCE_FRAME_LABEL

        return result

    @classmethod
    def from_dask_dataframe(cl, df, ensemble=None):
        """Returns a SourceFrame constructed from a Dask dataframe.

        Parameters
        ----------
        df: `dask.dataframe.DataFrame` or `list`
            a Dask dataframe to convert to a SourceFrame
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.SourceFrame`
            The constructed SourceFrame object.
        """
        # Create a SourceFrame by mapping the partitions to the appropriate meta, TapeSourceFrame
        # TODO(wbeebe@uw.edu): Determine if there is a better method
        result = df.map_partitions(TapeSourceFrame)
        result.ensemble = ensemble
        result.label = SOURCE_FRAME_LABEL
        return result


class ObjectFrame(EnsembleFrame):
    """A subclass of EnsembleFrame for Object data."""

    _partition_type = TapeObjectFrame  # Tracks the underlying data type

    def __init__(self, dsk, name, meta, divisions, ensemble=None):
        super().__init__(dsk, name, meta, divisions)
        self.label = OBJECT_FRAME_LABEL  # A label used by the Ensemble to identify this frame.
        self.ensemble = ensemble  # The Ensemble object containing this frame.

    @classmethod
    def from_parquet(
        cl,
        path,
        index=None,
        columns=None,
        ensemble=None,
    ):
        """Returns an ObjectFrame constructed from loading a parquet file.

        Parameters
        ----------
        path: `str` or `list`
            Source directory for data, or path(s) to individual parquet files. Prefix with a
            protocol like s3:// to read from alternative filesystems. To read from multiple
            files you can pass a globstring or a list of paths, with the caveat that they must all
            have the same protocol.
        columns: `str` or `list`, optional
            Field name(s) to read in as columns in the output. By default all non-index fields will
            be read (as determined by the pandas parquet metadata, if present). Provide a single
            field name instead of a list to read in the data as a Series.
        index: `str`, `list`, `False`, optional
            Field name(s) to use as the output frame index. Default is None and index will be
            inferred from the pandas parquet file metadata, if present. Use False to read all
            fields as columns.
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.ObjectFrame`
            The constructed ObjectFrame object.
        """
        # Read in the object Parquet file
        result = dd.read_parquet(
            path,
            index=index,
            columns=columns,
            split_row_groups=True,
            engine=TapeObjectArrowEngine,
        )
        result.ensemble = ensemble
        result.label = OBJECT_FRAME_LABEL

        return result

    @classmethod
    def from_dask_dataframe(cl, df, ensemble=None):
        """Returns an ObjectFrame constructed from a Dask dataframe.

        Parameters
        ----------
        df: `dask.dataframe.DataFrame` or `list`
            a Dask dataframe to convert to an ObjectFrame
        ensemble: `tape.ensemble.Ensemble`, optional
            A link to the Ensemble object that owns this frame.

        Returns
        ----------
        result: `tape.ObjectFrame`
            The constructed ObjectFrame object.
        """
        # Create an ObjectFrame by mapping the partitions to the appropriate meta, TapeObjectFrame
        # TODO(wbeebe@uw.edu): Determine if there is a better method
        result = df.map_partitions(TapeObjectFrame)
        result.ensemble = ensemble
        result.label = OBJECT_FRAME_LABEL
        return result


# Dask Dataframes are constructed indirectly using method dispatching and inference on the
# underlying data. So to ensure our subclasses behave correctly, we register the methods
# below.
#
# For more information, see https://docs.dask.org/en/latest/dataframe-extend.html
#
# The following should ensure that any Dask Dataframes which use TapeSeries or TapeFrames as their
# underlying data will be resolved as EnsembleFrames or EnsembleSeries as their parrallel
# counterparts. The underlying Dask Dataframe _meta will be a TapeSeries or TapeFrame.

get_parallel_type.register(TapeSeries, lambda _: EnsembleSeries)
get_parallel_type.register(TapeFrame, lambda _: EnsembleFrame)
get_parallel_type.register(TapeObjectFrame, lambda _: ObjectFrame)
get_parallel_type.register(TapeSourceFrame, lambda _: SourceFrame)


@make_meta_dispatch.register(TapeSeries)
def make_meta_series(x, index=None):
    # Create an empty TapeSeries to use as Dask's underlying object meta.
    result = x.head(0)
    return result


@make_meta_dispatch.register(TapeFrame)
def make_meta_frame(x, index=None):
    # Create an empty TapeFrame to use as Dask's underlying object meta.
    result = x.head(0)
    return result


@meta_nonempty.register(TapeSeries)
def _nonempty_tapeseries(x, index=None):
    # Construct a new TapeSeries with the same underlying data.
    data = _nonempty_series(x)
    return TapeSeries(data)


@meta_nonempty.register(TapeFrame)
def _nonempty_tapeseries(x, index=None):
    # Construct a new TapeFrame with the same underlying data.
    df = meta_nonempty_dataframe(x)
    return TapeFrame(df)


@make_meta_dispatch.register(TapeObjectFrame)
def make_meta_frame(x, index=None):
    # Create an empty TapeObjectFrame to use as Dask's underlying object meta.
    result = x.head(0)
    return result


@meta_nonempty.register(TapeObjectFrame)
def _nonempty_tapesourceframe(x, index=None):
    # Construct a new TapeObjectFrame with the same underlying data.
    df = meta_nonempty_dataframe(x)
    return TapeObjectFrame(df)


@make_meta_dispatch.register(TapeSourceFrame)
def make_meta_frame(x, index=None):
    # Create an empty TapeSourceFrame to use as Dask's underlying object meta.
    result = x.head(0)
    return result


@meta_nonempty.register(TapeSourceFrame)
def _nonempty_tapesourceframe(x, index=None):
    # Construct a new TapeSourceFrame with the same underlying data.
    df = meta_nonempty_dataframe(x)
    return TapeSourceFrame(df)



import pandas as pd

from tape.analysis.stetsonj import calc_stetson_J
from tape.analysis.structurefunction2 import calc_sf2


class TimeSeries:
    """Represent and analyze Rubin TimeSeries data"""

    def __init__(self, data=None):
        self.data = data
        self.meta = {"id": None}  # metadata dict
        self.colmap = {"time": None, "flux": None, "flux_err": None}  # column mapping

    # I/O
    def from_dict(
        self,
        data_dict,
        time_label="time",
        flux_label="flux",
        err_label="flux_err",
        band_label="band",
    ):
        """Build dataframe from a python dictionary

        Parameters
        ----------
        data_dict : `dict`
            Dictionary contaning the data.
        time_label: `str`
            Name for column containing time information.
        flux_label: `str`
            Name for column containing signal
            (flux, magnitude, etc) information.
        err_label: `str`
            Name for column containing error information.
        band_label: `str`
            Name for column containing filter information.
        """

        try:
            data_dict[band_label]
        except KeyError as exc:
            raise KeyError(f"The indicated label '{band_label}' was not found.") from exc
        index = self._build_index(data_dict[band_label])
        data_dict = {key: data_dict[key] for key in data_dict if key != band_label}
        self.data = pd.DataFrame(data=data_dict, index=index).sort_index()

        labels = [time_label, flux_label, err_label]
        for label, quantity in zip(labels, list(self.colmap.keys())):
            if (quantity == "flux_err") and (label is None):  # flux_err is optional
                continue

            if label in self.data.columns:
                self.colmap[quantity] = label
            else:
                raise KeyError(f"The indicated label '{label}' was not found.")

        return self

    def dropna(self, **kwargs):
        """Handle NaN values, wrapper for pandas.DataFrame.dropna"""
        self.data = self.data.dropna(**kwargs)
        return self

    def from_dataframe(
        self, data, object_id, time_label="time", flux_label="flux", err_label="flux_err", band_label="band"
    ):
        """Loader function for inputing data from a dataframe.

        Parameters
        ----------
        data : `pandas.DataFrame`
            The data for the time serires.
        object_id : `str`
            The ID of the current object.
        time_label: `str`
            Name for column containing time information.
        flux_label: `str`
            Name for column containing signal
            (flux, magnitude, etc) information.
        err_label: `str`
            Name for column containing error information.
        band_label: `str`
            Name for column containing filter information.
        """
        self.data = data
        self.meta["id"] = object_id

        # Index the timeseries on band.
        index = self._build_index(self.data[band_label])
        self.data.index = index

        labels = [time_label, flux_label, err_label]
        for label, quantity in zip(labels, list(self.colmap.keys())):
            if (quantity == "flux_err") and (label is None):  # flux_err is optional
                continue

            if label in self.data.columns:
                self.colmap[quantity] = label
            else:
                raise KeyError(f"The indicated label '{label}' was not found.")

        return self

    @property
    def time(self):
        """Time values stored as a Pandas Series"""
        return self.data[self.colmap["time"]]

    @property
    def flux(self):
        """Flux values stored as a Pandas Series"""
        return self.data[self.colmap["flux"]]

    @property
    def flux_err(self):
        """Flux error values stored as a Pandas Series"""
        if self.colmap["flux_err"] is not None:  # Errors are not mandatory
            return self.data[self.colmap["flux_err"]]
        return None

    @property
    def band(self):
        """Band labels stored as a Pandas Index"""
        return self.data.index.get_level_values("band")

    def _build_index(self, band):
        """Build pandas multiindex from band array"""
        count_dict = {}
        idx = []
        for b in band:
            count = count_dict.get(b, 0)
            idx.append(count)

            # Increment count for this band or insert 1 there wasn't an ongoing count.
            count_dict[b] = count + 1
        tuples = zip(band, idx)
        index = pd.MultiIndex.from_tuples(tuples, names=["band", "index"])
        return index

    def stetson_J(self, band=None):
        """Compute the stetsonJ statistic on data from one or several bands

        Parameters
        ----------
        band : `str` or `list` of `str`
            Single band descriptor, or list of such descriptors.

        Returns
        -------
        stetsonJ : `dict`
            StetsonJ statistic for each of input bands.

        Note
        ----------
        In case that no value for band is passed, the function is executed
        on all available bands.
        """
        return calc_stetson_J(self.flux, self.flux_err, self.band, band_to_calc=band)

    def sf2(self, sf_method="basic", argument_container=None):
        """Compute the structure function squared statistic on data

        Parameters
        ----------
        bins : `numpy.array` or `list`
            Manually provided bins, if not provided then bins are computed using
            the `method` kwarg
        band_to_calc : `str` or `list` of `str`
            Single band descriptor, or list of such descriptors.
        method : 'str'
            The binning method to apply, choices of 'size'; which seeks an even
            distribution of samples per bin using quantiles, 'length'; which
            creates bins of equal length in time and 'loglength'; which creates
            bins of equal length in log time.
        sthresh : 'int'
            Target number of samples per bin.

        Returns
        -------
        stetsonJ : `dict`
            Structure function squared statistic for each of input bands.

        Note
        ----------
        In case that no value for band_to_calc is passed, the function is executed
        on all available bands.
        """
        if self.meta["id"]:
            lc_id = [self.meta["id"]] * len(self.time)
        else:
            lc_id = [0] * len(self.time)
        return calc_sf2(
            time=self.time,
            flux=self.flux,
            err=self.flux_err,
            band=self.band,
            lc_id=lc_id,
            sf_method=sf_method,
            argument_container=argument_container,
        )



import numpy as np
import pytest

from tape.analysis.structure_function.sf_light_curve import StructureFunctionLightCurve


def test_lightcurve_creation():
    """Test the creation of a lightcurve, check that the number of
    expected delta values equals expectations.
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    expected_length = test_t.size * (test_t.size - 1) / 2
    assert expected_length == sf_lc._all_d_times.size


def test_lightcurve_sample_equals_all_by_default():
    """When we do not request a sub set of delta values, we expect
    that the `._all_*` arrays will be the same as the `.sample_*`
    arrays.
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    assert sf_lc._all_d_times.size == sf_lc.sample_d_times.size
    assert sf_lc._all_d_fluxes.size == sf_lc.sample_d_fluxes.size
    assert sf_lc._all_sum_squared_error.size == sf_lc.sample_sum_squared_error.size


def test_lightcurve_get_sub_sample():
    """Make sure that sample values have the length requested"""
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    num_samples_requested = 5
    sf_lc.select_difference_samples(num_samples_requested)

    assert sf_lc.sample_d_times.size == num_samples_requested


def test_lightcurve_get_sub_sample_raises():
    """Make sure that the method raises an exception when we request more samples
    than are available."""
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    with pytest.raises(ValueError) as excinfo:
        num_samples_requested = 50000
        _ = sf_lc.select_difference_samples(num_samples_requested)

    assert f"Requesting {num_samples_requested} samples, but only" in str(excinfo.value)


def test_lightcurve_get_sub_sample_with_duplicated_generator():
    """Make sure that the input random generator is respected, and will
    produce repeatable results when reset with the same seed.
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    num_samples_requested = 5
    random_seed = 13

    rng = np.random.default_rng(random_seed)
    sf_lc.select_difference_samples(num_samples_requested, random_generator=rng)
    output_1 = sf_lc.sample_d_fluxes

    rng = np.random.default_rng(random_seed)
    sf_lc.select_difference_samples(num_samples_requested, random_generator=rng)
    output_2 = sf_lc.sample_d_fluxes

    assert np.all(output_1 == output_2)


def test_lightcurve_get_sub_sample_with_same_generator():
    """Make sure the input random generator is respected and produces distinct
    result when run multiple times.
    """
    test_t = np.array([1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2])
    test_y = np.array([0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2])
    test_yerr = np.array([0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02])

    sf_lc = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    num_samples_requested = 5
    random_seed = 13
    rng = np.random.default_rng(random_seed)

    sf_lc.select_difference_samples(num_samples_requested, random_generator=rng)
    output_1 = sf_lc.sample_d_fluxes

    sf_lc.select_difference_samples(num_samples_requested, random_generator=rng)
    output_2 = sf_lc.sample_d_fluxes

    assert not np.all(output_1 == output_2)


def test_lightcurve_validate_sufficient_observations():
    """Make sure that we raise an exception if there aren't enough
    observations to calculate Structure Function differences.
    """
    test_t = np.array([1.11])
    test_y = np.array([0.11])
    test_yerr = np.array([0.1])

    with pytest.raises(ValueError) as excinfo:
        _ = StructureFunctionLightCurve(test_t, test_y, test_yerr)

    assert "Too few observations provided to create `LightCurve`." in str(excinfo.value)



"""Test feature extraction with light_curve package"""

import light_curve as licu
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

from tape import Ensemble
from tape.analysis.feature_extractor import FeatureExtractor
from tape.utils import ColumnMapper


def test_stetsonk():
    stetson_k = licu.StetsonK()

    time = np.array([5.0, 4.0, 3.0, 2.0, 1.0, 0.0] * 2)
    flux = 1.0 + time**2.0
    err = np.full_like(time, 0.1, dtype=np.float32)
    band = np.r_[["g"] * 6, ["r"] * 6]

    extract_features = FeatureExtractor(stetson_k)
    result = extract_features(time=time, flux=flux, err=err, band=band, band_to_calc="g")
    assert result.shape == (1,)
    assert_array_equal(result.index, ["stetson_K"])
    assert_allclose(result.values, 0.84932, rtol=1e-5)
    assert_array_equal(result.dtypes, np.float64)


def test_multiple_features_with_ensemble():
    n = 5

    object1 = {
        "id": np.full(n, 1),
        "time": np.arange(n, dtype=np.float64),
        "flux": np.linspace(1.0, 2.0, n),
        "err": np.full(n, 0.1),
        "band": np.full(n, "g"),
    }
    object2 = {
        "id": np.full(2 * n, 2),
        "time": np.arange(2 * n, dtype=np.float64),
        "flux": np.r_[np.linspace(1.0, 2.0, n), np.linspace(1.0, 2.0, n)],
        "err": np.full(2 * n, 0.01),
        "band": np.r_[np.full(n, "g"), np.full(n, "r")],
    }
    rows = {column: np.concatenate([object1[column], object2[column]]) for column in object1}

    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")
    ens = Ensemble(client=False).from_source_dict(rows, cmap)

    extractor = licu.Extractor(licu.AndersonDarlingNormal(), licu.InterPercentileRange(0.25), licu.StetsonK())
    result = ens.batch(
        extractor,
        band_to_calc="g",
    ).compute()

    assert result.shape == (2, 3)
    assert_array_equal(result.columns, ["anderson_darling_normal", "inter_percentile_range_25", "stetson_K"])
    assert_allclose(result, [[0.114875, 0.625, 0.848528]] * 2, atol=1e-5)


def test_otsu_with_ensemble_all_bands():
    n = 10
    assert n % 2 == 0

    object1 = {
        "id": np.full(n, 1),
        "time": np.arange(n, dtype=np.float64),
        "flux": np.r_[np.zeros(n // 2), np.ones(n // 2)],
        "err": np.full(n, 0.1),
        "band": np.full(n, "g"),
    }
    object2 = {
        "id": np.full(n, 2),
        "time": object1["time"],
        "flux": object1["flux"],
        "err": object1["err"],
        "band": np.r_[np.full(n // 2, "g"), np.full(n // 2, "r")],
    }
    rows = {column: np.concatenate([object1[column], object2[column]]) for column in object1}

    cmap = ColumnMapper(id_col="id", time_col="time", flux_col="flux", err_col="err", band_col="band")
    ens = Ensemble(client=False).from_source_dict(rows, cmap)

    result = ens.batch(
        licu.OtsuSplit(),
        band_to_calc=None,
    ).compute()

    assert result.shape == (2, 4)
    assert_array_equal(
        result.columns, ["otsu_mean_diff", "otsu_std_lower", "otsu_std_upper", "otsu_lower_to_all_ratio"]
    )
    assert_allclose(result, [[1.0, 0.0, 0.0, 0.5]] * 2, atol=1e-5)



from typing import Dict, List

import numpy as np
import pandas as pd

from tape.analysis.base import AnalysisFunction
from tape.analysis.structure_function import SF_METHODS
from tape.analysis.structure_function.sf_light_curve import StructureFunctionLightCurve


__all__ = ["StructureFunction2", "calc_sf2"]


class StructureFunction2(AnalysisFunction):
    """Calculate structure function squared"""

    def cols(self, ens: "Ensemble") -> List[str]:
        return [
            ens._time_col,
            ens._flux_col,
            ens._err_col,
            ens._band_col,
            ens._id_col,
        ]

    def meta(self, ens: "Ensemble") -> Dict[str, type]:
        return {"lc_id": int, "band": str, "dt": float, "sf2": float, "1_sigma": float}

    def on(self, ens: "Ensemble") -> List[str]:
        return [ens._id_col]

    def __call__(
        self, time, flux, err=None, band=None, lc_id=None, *, sf_method="basic", argument_container=None
    ) -> pd.DataFrame:
        """Calculate structure function squared using one of a variety of structure
        function calculation methods defined by the input argument `sf_method`, or
        in the argument container object.


        Parameters
        ----------
        time : `numpy.ndarray` (N,) or `None`
            Array of times when measurements were taken. If all array values are
            `None` or if a scalar `None` is provided, then equidistant time between
            measurements is assumed.
        flux : `numpy.ndarray` (N,)
            Array of flux/magnitude measurements.
        err : `numpy.ndarray` (N,), `float`, or `None`, optional
            Array of associated flux/magnitude errors. If a scalar value is provided
            we assume that error for all measurements. If `None` is provided, we
            assume all errors are 0. By default None
        band : `numpy.ndarray` (N,), optional
            Array of associated band labels, by default None
        lc_id : `numpy.ndarray` (N,), optional
            Array of lightcurve ids per data point. By default None
        sf_method : str, optional
            The structure function calculation method to be used, by default "basic".
        argument_container : StructureFunctionArgumentContainer, optional
            Container object for additional configuration options, by default None.

        Returns
        -------
        sf2 : `pandas.DataFrame`
            Structure function squared for each of input bands.

        Notes
        ----------
        In case that no value for `band_to_calc` is passed, the function is
        executed on all available bands in `band`.
        """

        argument_container = _create_arg_container_if_needed(sf_method, argument_container)

        band = _validate_band(band, flux, argument_container)

        lc_id = _validate_lightcurve_id(lc_id, flux, argument_container)

        sf_method = _validate_sf_method(sf_method, argument_container)

        _validate_argument_container(argument_container)

        unq_band = np.unique(band)
        unq_ids = np.unique(lc_id)

        band_to_calc = argument_container.band_to_calc
        if band_to_calc is None:
            band_to_calc = unq_band
        if isinstance(band_to_calc, str):
            band_to_calc = [band_to_calc]

        assert hasattr(band_to_calc, "__iter__") is True

        ids = []
        dts = []
        bands = []
        sf2s = []
        sf2_err = []
        sf2_lower_error = []
        sf2_upper_error = []
        for b in band_to_calc:
            if b in unq_band:
                band_mask = band == b

                # Mask on band
                times = _extract_time(time=time, band_mask=band_mask, argument_container=argument_container)

                errors = _extract_error(err=err, band_mask=band_mask)

                fluxes = np.array(flux)[band_mask]
                lc_ids = np.array(lc_id)[band_mask]

                # Create stacks of critical quantities, indexed by id
                id_masks = [lc_ids == lc for lc in unq_ids]

                lightcurves = []
                for mask in id_masks:
                    try:
                        sf_lc = StructureFunctionLightCurve(
                            times=times[mask], fluxes=fluxes[mask], errors=errors[mask]
                        )
                        lightcurves.append(sf_lc)
                    except ValueError:
                        # Exception raised by StructureFunctionLightCurve when there are too few data point.
                        print("Attempted to create a Lightcurve with too few data points.")

                if len(lightcurves):
                    sf_calculator = SF_METHODS[sf_method](lightcurves, argument_container)

                    # `aggregated_dts` and `aggregated_sfs` will have the shape:
                    # [calc_rep(0:arg_container.calc_repetitions)][lc_id(0:num_lightcurves)][bin(0:num_dt_bins)]
                    aggregated_dts: List[np.ndarray] = []
                    aggregated_sfs: List[np.ndarray] = []
                    rng = np.random.default_rng(argument_container.random_seed)
                    for _ in range(argument_container.calculation_repetitions):
                        if argument_container.estimate_err:
                            sf_calculator._bootstrap(random_generator=rng)

                        tmp_dts, tmp_sfs = sf_calculator.calculate()
                        aggregated_dts.append(tmp_dts)
                        aggregated_sfs.append(tmp_sfs)

                    # find the median value for each (lightcurve, dt_bin) coordinate
                    res_dts = np.nanmedian(aggregated_dts, axis=0)
                    res_sfs = np.nanmedian(aggregated_sfs, axis=0)

                    if _no_results_found(aggregated_sfs):
                        res_err = np.zeros_like(res_sfs)
                        lower_error = np.zeros_like(res_err)
                        upper_error = np.zeros_like(res_err)
                    else:
                        # Subtract the upper and lower quantiles and remove the outer
                        # axis that has length 1. The resulting shape will be the same
                        # as `res_dts`` and `res_sfs`.
                        lower_quantile, upper_quantile = np.nanquantile(
                            aggregated_sfs,
                            (
                                argument_container.lower_error_quantile,
                                argument_container.upper_error_quantile,
                            ),
                            axis=0,
                        )

                        res_err = (upper_quantile - lower_quantile) / 2
                        lower_error = res_sfs - lower_quantile
                        upper_error = upper_quantile - res_sfs

                    res_ids = [[str(unq_ids[i])] * len(arr) for i, arr in enumerate(res_dts)]
                    res_bands = [[b] * len(arr) for arr in res_dts]

                    ids.append(np.hstack(res_ids))
                    bands.append(np.hstack(res_bands))
                    dts.append(np.hstack(res_dts))
                    sf2s.append(np.hstack(res_sfs))
                    sf2_err.append(np.hstack(res_err))
                    if argument_container.report_upper_lower_error_separately:
                        sf2_lower_error.append(np.hstack(lower_error))
                        sf2_upper_error.append(np.hstack(upper_error))

        id_stack = []
        band_stack = []
        dts_stack = []
        sf2_stack = []
        sigma_stack = []
        sf2_lower_stack = []
        sf2_upper_stack = []

        if len(ids):
            id_stack = np.hstack(ids)
            if argument_container.combine:
                id_stack = ["combined"] * len(id_stack)

            band_stack = np.hstack(bands)
            dts_stack = np.hstack(dts)
            sf2_stack = np.hstack(sf2s)
            sigma_stack = np.hstack(sf2_err)

            if argument_container.report_upper_lower_error_separately:
                sf2_lower_stack = np.hstack(sf2_lower_error)
                sf2_upper_stack = np.hstack(sf2_upper_error)

        data_frame_dict = {
            "lc_id": id_stack,
            "band": band_stack,
            "dt": dts_stack,
            "sf2": sf2_stack,
            "1_sigma": sigma_stack,
        }

        if argument_container.report_upper_lower_error_separately:
            data_frame_dict["lower_error"] = sf2_lower_stack
            data_frame_dict["upper_error"] = sf2_upper_stack

        sf2_df = pd.DataFrame(data_frame_dict)
        return sf2_df


calc_sf2 = StructureFunction2()
calc_sf2.__doc__ = StructureFunction2.__call__.__doc__


def _create_arg_container_if_needed(sf_method, argument_container):
    """If no argument container was provided, we'll create a default container
    with the default values using the value of `sf_method` to determine the correct
    type.

    Parameters
    ----------
    sf_method : str
        The structure function calculation method to be used.
    argument_container : StructureFunctionArgumentContainer
        Container object for additional configuration options.

    Returns
    -------
    StructureFunctionArgumentContainer
        Container object for additional configuration options.
    """

    if argument_container is None:
        argument_container_type = SF_METHODS[sf_method].expected_argument_container()
        argument_container = argument_container_type()
        argument_container.sf_method = sf_method

    return argument_container


def _validate_band(band, flux, argument_container):
    """The argument `band` can be set as an input argument or inside the
    `argument_container`. If `band` is provided in the function call with
    non-default values, we'll use those. Otherwise, we'll look inside
    `argument_container` and use the values found there. If we still can't find
    non-`None` values, we'll create a minimal fallback array of 1's with length
    equal to the input `flux` array.

    Parameters
    ----------
    band : `numpy.ndarray` (N,)
        Array of associated band labels.
    flux : `numpy.ndarray` (N,)
        Array of flux/magnitude measurements.
    argument_container : StructureFunctionArgumentContainer
        Container object for additional configuration options.

    Returns
    -------
    `numpy.ndarray` (N,)
        The band information to be used for structure function calculations.

    Raises
    ------
    ValueError
        If the provided band and flux arrays have different lengths, raise an
        exception.
    """

    if band is None:
        band = argument_container.band

    # if band is still `None`, create a fake array of bands. We'll use numpy int8
    # values to use as little memory as possible.
    if band is None:
        band = np.zeros(len(flux), dtype=np.int8)

    if len(band) != len(flux):
        raise ValueError(
            "Value of `band` should be `None` or array with the same length as the `flux` array."
        )

    return band


def _validate_lightcurve_id(lc_id, flux, argument_container):
    """The argument `lc_id` can be set as an input argument or inside the
    `argument_container`. If `lc_id` is provided in the function call with
    non-default values, we'll use those. Otherwise, we'll look inside
    `argument_container` and use the values found there. If we still can't find
    non-`None` values, we'll create a minimal fallback array of 1's with length
    equal to the input `flux` array.

    Parameters
    ----------
    lc_id : `numpy.ndarray` (N,), optional
        Array of lightcurve ids per data point.
    flux : `numpy.ndarray` (N,)
        Array of flux/magnitude measurements.
    argument_container : StructureFunctionArgumentContainer
        Container object for additional configuration options.

    Returns
    -------
    `numpy.ndarray` (N,)
        The lightcurve id information to be used for structure function
        calculations.

    Raises
    ------
    ValueError
        If the provided lc_id and flux arrays have different lengths, raise an
        exception.
    """

    if lc_id is None:
        lc_id = argument_container.lc_id

    # if the light curve value is still `None`, create an array of all 0's.
    # Using numpy.int8 values to save memory.
    if lc_id is None:
        lc_id = np.zeros(len(flux), dtype=np.int8)

    if len(lc_id) != len(flux):
        raise ValueError(
            "Value of `lc_id` should be `None` or array with the same length as the `flux` array."
        )

    return lc_id


def _validate_sf_method(sf_method, argument_container):
    """The argument `sf_method` can be set as an input argument or inside the
    `argument_container`. If `sf_method` is provided in the function call with
    a non-default value, we'll use that. If the default value for `sf_method`
    ("basic") has been provided, we'll check `argument_container` to see if the
    user provided a different `sf_method` value there.

    Parameters
    ----------
    sf_method : str, optional
        The structure function calculation method to be used, by default "basic".
    argument_container : StructureFunctionArgumentContainer
        Container object for additional configuration options.

    Returns
    -------
    `str`
        The band information to be used for structure function calculations.
    """

    if sf_method == "basic":
        sf_method = argument_container.sf_method

    if sf_method not in SF_METHODS:
        raise ValueError("Unknown structure function calculator method: " + str(sf_method))

    return sf_method


def _validate_argument_container(argument_container):
    """Raise an exception if this type of argument container does not match the
    type of argument container required by the structure function calculation
    method
    """

    if type(argument_container) is not SF_METHODS[argument_container.sf_method].expected_argument_container():
        raise TypeError("Argument container does not match Structure Function calculator method")


def _extract_time(time, band_mask, argument_container):
    """This will process the input time values, apply a band mask and return a
    numpy array of timestamps.

    Parameters
    ----------
    time : `numpy.ndarray` (N,) or `None`
        Array of times when measurements were taken. If all array values are
        `None` or if a scalar `None` is provided, then equidistant time between
        measurements is assumed.
    band_mask : `numpy mask` (N,)
        Array of associated band labels, by default None
    argument_container : StructureFunctionArgumentContainer, optional
        Container object for additional configuration options, by default None.
    """

    times = None

    # if the user passed in a scalar `None` value, or if explicitly ignoring
    # timestamps using the argument_container flag, create a numpy array
    # with a single `None` element. Otherwise assume the user passed an
    # array of timestamps to be masked with `band_mask`.
    # Note: some or all timestamps could be `None`.
    if time is None or argument_container.ignore_timestamps:
        times = np.array(None)
    else:
        times = np.array(time)[band_mask]

    # if all elements in `times` are `None`, we assume equidistant times
    # between measurements. To do so, we'll create an array of integers
    # from 0 to N-1 where N is the number of flux values for this band.
    if np.all(np.equal(times, None)):
        times = np.arange(sum(band_mask), dtype=int)

    return times


def _extract_error(err, band_mask):
    """Process the input err value(s), apply a band_mask, and return a numpy
    array of error values.

    Parameters
    ----------
    err : `numpy.ndarray` (N,), `float`, or `None`
        Array of associated flux/magnitude errors. If a scalar value is provided
        we assume that error for all measurements. If `None` is provided, we
        assume all errors are 0. By default None
    band_mask : `numpy mask` (N,)
        Array of associated band labels, by default None
    """

    errors = None
    # assume all errors are 0 if `None` is provided
    if err is None:
        errors = np.zeros(sum(band_mask))

    # assume the same error for all measurements if a scalar value is
    # provided
    elif np.isscalar(err):
        errors = np.ones(sum(band_mask)) * err

    # otherwise assume one error value per measurement
    else:
        errors = np.array(err)[band_mask]

    return errors


def _no_results_found(aggregated_sfs: List[np.ndarray]) -> bool:
    """This helper function determines if there are results from calculating
    the Structure Function for a given input. An example instances where there
    might not be results would be calculating the SF for a Lightcurve object
    that contains only 1 observation.

    Parameters
    ----------
    aggregated_sfs : List[np.ndarray]
        The output from the Structure Function calculation. This is a 3 dimensional
        array with dimensions [calculation_repetitions][lightcurve_id][sf_result]
        where calculation_repetition has range (0:arg_container.calc_repetitions)
        lightcurve_id has range (0:num_lightcurves)
        and sf_result has range (0:num_dt_bins)

    Returns
    -------
    bool
        True if no results were found.
    """

    no_results_found = False

    # `shape[2]==0` implies that there wasn't enough data to perform the SF calculation
    # For instance, only 1 observation.
    if np.shape(aggregated_sfs)[2] == 0:
        no_results_found = True

    return no_results_found



import numpy as np

from tape.analysis.structure_function.base_argument_container import StructureFunctionArgumentContainer
from tape.analysis.structure_function.base_calculator import StructureFunctionCalculator

# MacLeod et al. 2012, Erratum 2014ApJ...782..119M
CONVERSION_TO_SIGMA = 0.74


class Macleod2012StructureFunctionCalculator(StructureFunctionCalculator):
    """This class implements the structure function calculation described in
    MacLeod et al. 2012, 2012ApJ...753..106M [https://arxiv.org/abs/1112.0679]
    MacLeod et al. 2012, Erratum 2014ApJ...782..119M

    `SF_obs(deltaT) = 0.74 * IQR`

    Where `IQR` is the interquartile range between 25% and 75% of the sorted
    (y(t) - y(t+delta_t)) distribution.

    Note that the return value is structure function squared.

    Additional references:
    Kozlowski 2016, 2016ApJ...826..118K [https://arxiv.org/abs/1604.05858]
    """

    def calculate(self):
        dts, sfs = self._calculate_binned_statistics(statistic_to_apply=self.calculate_iqr_sf2_statistic)

        return dts, sfs

    def calculate_iqr_sf2_statistic(self, input):
        """For a given set of binned metrics (in this case delta fluxes) calculate
        the interquartile range.

        Parameters
        ----------
        input : `np.ndarray` (N,)
            The delta flux values that correspond to a given delta time bin.

        Returns
        -------
        float
            Result of calculation defined in MacLeod et al. 2012, Erratum 2014ApJ...782..119M:

            `SF(dt) = 0.74 * IQR`
        """
        # calculate interquartile range between 25% and 75%.
        iqr = np.subtract(*np.percentile(input, [75, 25]))

        return (CONVERSION_TO_SIGMA * iqr) ** 2

    @staticmethod
    def name_id() -> str:
        return "macleod_2012"

    @staticmethod
    def expected_argument_container() -> type:
        return StructureFunctionArgumentContainer



import numpy as np
import pytest

from tape.analysis.structurefunction2 import calc_sf2
from tape.analysis.structure_function.bauer_2009b.calculator import (
    Bauer2009BStructureFunctionCalculator,
)
from tape.analysis.structure_function.base_argument_container import StructureFunctionArgumentContainer
from tape.analysis.structure_function.sf_light_curve import StructureFunctionLightCurve


def test_basic_calculation():
    """Most basic test possible. Inputs are expected to be 2d numpy arrays."""

    test_t = np.atleast_2d([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    test_y = np.atleast_2d([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    test_e = np.atleast_2d([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    test_lc = [StructureFunctionLightCurve(test_t, test_y, test_e)]
    arg_container = StructureFunctionArgumentContainer()
    arg_container.combine = False

    sf_calculator = Bauer2009BStructureFunctionCalculator(test_lc, arg_container)

    res = sf_calculator.calculate()

    assert res


def test_sf2_base_case_bauer_2009b():
    """
    Base test case accessing calc_sf2 directly. Uses `Bauer 2009b` SF calculation method.
    """
    lc_id = [1, 1, 1, 1, 1, 1, 1, 1]
    test_t = [1.11, 2.23, 3.45, 4.01, 5.67, 6.32, 7.88, 8.2]
    test_y = [0.11, 0.23, 0.45, 0.01, 0.67, 0.32, 0.88, 0.2]
    test_yerr = [0.1, 0.023, 0.045, 0.1, 0.067, 0.032, 0.8, 0.02]
    test_band = np.array(["r"] * len(test_y))
    test_sf_method = "bauer_2009b"

    res = calc_sf2(
        time=test_t, flux=test_y, err=test_yerr, band=test_band, lc_id=lc_id, sf_method=test_sf_method
    )

    assert res["dt"][0] == pytest.approx(3.1482, rel=0.001)
    assert res["sf2"][0] == pytest.approx(0.0225667, rel=0.001)



from typing import Optional
import numpy as np

from tape.analysis.light_curve import LightCurve

MIN_OBSERVATIONS_REQUIRED_FOR_SF = 3


class StructureFunctionLightCurve(LightCurve):
    def __init__(self, times: np.ndarray, fluxes: np.ndarray, errors: np.ndarray):
        super().__init__(times, fluxes, errors, MIN_OBSERVATIONS_REQUIRED_FOR_SF)

        # The derived data from the primary data
        self._all_d_times: np.array = []
        self._all_d_fluxes: np.array = []
        self._all_sum_squared_error: np.array = []

        # _public_ derived data
        self.sample_d_times: np.array = []
        self.sample_d_fluxes: np.array = []
        self.sample_sum_squared_error: np.array = []

        self.number_of_difference_values = None

        self._calculate_differences()

    def _calculate_differences(self):
        """Calculate the difference between all possible pairs of time and flux.
        Also calculate the sum of all possible pairs of error^2. To avoid
        duplicate values, we filter out any differences that correspond to a
        time difference less than 0.
        """
        dt_matrix = self._times.reshape((1, self._times.size)) - self._times.reshape((self._times.size, 1))
        self._all_d_times = dt_matrix[dt_matrix > 0].flatten()
        self.sample_d_times = self._all_d_times
        self.number_of_difference_values = self._all_d_times.size

        df_matrix = self._fluxes.reshape((1, self._fluxes.size)) - self._fluxes.reshape(
            (self._fluxes.size, 1)
        )
        self._all_d_fluxes = df_matrix[dt_matrix > 0].flatten()
        self.sample_d_fluxes = self._all_d_fluxes

        err2_matrix = (
            self._errors.reshape((1, self._errors.size)) ** 2
            + self._errors.reshape((self._errors.size, 1)) ** 2
        )
        self._all_sum_squared_error = err2_matrix[dt_matrix > 0].flatten()
        self.sample_sum_squared_error = self._all_sum_squared_error

    def select_difference_samples(
        self, number_of_samples: Optional[int] = None, random_generator: Optional[np.random.Generator] = None
    ):
        """Take a random sample of time and flux differences and the sum of squared
        errors. The samples are selected without replacement. The resulting
        sub-sample is not guaranteed to have the same order as the input
        differences.

        Parameters
        ----------
        number_of_samples : int, optional
            Defines the number of samples to be randomly selected from the total
            number of difference values. If not specified, take all of the
            avaliable values
        random_generator: np.random.Generator, optional
            A Numpy random.Generator to sample the lightcurve difference. This
            allows for repeatable random samples to be selected. By default None.

        Raises
        ------
        ValueError
            If samples are requested than are present in the light curve raise
            ValueError.

        """

        if number_of_samples is None:
            number_of_samples = self.number_of_difference_values

        if number_of_samples > self.number_of_difference_values:
            raise ValueError(
                f"Requesting {number_of_samples} samples, but only "
                f"{self.number_of_difference_values} are present in the lightcurve"
            )

        # Initialize a random generator if one was not provided
        if random_generator is None:
            random_generator = np.random.default_rng()

        # Stack the time and flux differences and errors.
        data_block = np.vstack((self._all_d_times, self._all_d_fluxes, self._all_sum_squared_error)).astype(
            float
        )

        # Randomly choose `number_of_samples` from the data_block. Return the
        # random samples into distinct arrays.
        self.sample_d_times, self.sample_d_fluxes, self.sample_sum_squared_error = random_generator.choice(
            data_block, number_of_samples, replace=True, axis=1, shuffle=False
        )



import numpy as np

from tape.analysis.structure_function.base_argument_container import StructureFunctionArgumentContainer
from tape.analysis.structure_function.base_calculator import StructureFunctionCalculator

PI_OVER_2 = np.pi / 2.0


class Bauer2009BStructureFunctionCalculator(StructureFunctionCalculator):
    """This class implements the structure function calculation described in
    Eqn. 5 of Bauer et al. 2009, 2009ApJ...696.1241B [https://arxiv.org/abs/0902.4103]

    `SF(tau) = sqrt( (pi/2) * mean(abs(delta_flux))^2 - mean(err^2) )`

    Note that the return value is structure function squared.

    Additional references:
    Graham et al. 2014MNRAS.439..703G [https://arxiv.org/abs/1401.1785]
    """

    def calculate(self):
        # gather the means of the abs(delta_fluxes) per bin
        values_to_be_binned = [np.abs(lc.sample_d_fluxes) for lc in self._lightcurves]
        dts, mean_d_flux_per_bin = self._calculate_binned_statistics(sample_values=values_to_be_binned)

        # gather the means of the sigma^2 (error) per bin
        values_to_be_binned = [lc.sample_sum_squared_error for lc in self._lightcurves]
        _, mean_err2_per_bin = self._calculate_binned_statistics(sample_values=values_to_be_binned)

        # calculate the structure function
        sfs = PI_OVER_2 * np.square(mean_d_flux_per_bin) - mean_err2_per_bin

        return dts, sfs

    @staticmethod
    def name_id() -> str:
        return "bauer_2009b"

    @staticmethod
    def expected_argument_container() -> type:
        return StructureFunctionArgumentContainer



class IndexCallable:
    """Provide getitem syntax for functions

    >>> def inc(x):
    ...     return x + 1

    >>> I = IndexCallable(inc)
    >>> I[3]
    4
    """

    __slots__ = ("fn", "dirty", "ensemble")

    def __init__(self, fn, dirty, ensemble):
        self.fn = fn
        self.dirty = dirty  # propagate metadata
        self.ensemble = ensemble  # propagate ensemble metadata

    def __getitem__(self, key):
        result = self.fn(key)
        result.dirty = self.dirty  # propagate metadata
        result.ensemble = self.ensemble  # propagate ensemble metadata
        return result



from tape.analysis.structure_function.base_calculator import StructureFunctionCalculator


def register_sf_subclasses():
    """This method will identify all of the subclasses of `StructureFunctionCalculator`
    and build a dictionary that maps `name : subclass`.

    Returns
    -------
    dict
        A dictionary of all of subclasses of `StructureFunctionCalculator`. Where
        the str returned from `subclass.name_id()` is the key, and the class is
        the value.

    Raises
    ------
    ValueError
        If a duplicate key is found, a ValueError will be raised. This would
        likely occur if a user copy/pasted an existing subclass but failed to
        update the unique name_id string.
    """
    subclass_dict = {}
    for subcls in StructureFunctionCalculator.__subclasses__():
        if subcls.name_id() in subclass_dict:
            raise ValueError(
                "Attempted to add duplicate Structure Function calculator name to SF_METHODS: "
                + str(subcls.name_id())
            )

        subclass_dict[subcls.name_id()] = subcls

    return subclass_dict


def update_sf_subclasses():
    """This function is used to register newly created subclasses of the
    `StructureFunctionCalculator`.
    """
    for subcls in StructureFunctionCalculator.__subclasses__():
        if subcls.name_id() not in SF_METHODS.keys():
            SF_METHODS[subcls.name_id()] = subcls


# The dictionary of all available subclasses of the StructureFunctionCalculator.
SF_METHODS = register_sf_subclasses()



from abc import ABC, abstractmethod
from typing import List

import numpy as np
from scipy.stats import binned_statistic

from tape.analysis.structure_function.base_argument_container import StructureFunctionArgumentContainer
from tape.analysis.structure_function.sf_light_curve import StructureFunctionLightCurve


class StructureFunctionCalculator(ABC):
    """This is the base class from which all other Structure Function calculator
    methods inherit. Extend this class if you want to create a new Structure
    Function calculation method.
    """

    def __init__(
        self,
        lightcurves: List[StructureFunctionLightCurve],
        argument_container: StructureFunctionArgumentContainer,
    ):
        self._lightcurves = lightcurves
        self._argument_container = argument_container

        self._bins = argument_container.bins  # defaults to None
        self._binning_method = argument_container.bin_method
        self._bin_count_target = argument_container.bin_count_target
        self._equally_weight_lightcurves = argument_container.equally_weight_lightcurves
        self._dts = []
        self._all_d_fluxes = []
        self._sum_error_squared = []
        self._difference_values_per_lightcurve: List[int] = []
        return

    @abstractmethod
    def calculate(self):
        """Abstract method that must be implemented by the child class."""
        raise (NotImplementedError, "Must be implemented by the child class")

    def _bootstrap(self, random_generator=None):
        """This method creates the boostraped samples of difference values"""
        self._get_difference_values_per_lightcurve()

        # if the user defined equal weight in the argument container,
        # use that, otherwise, go to specified number of difference values.
        if self._equally_weight_lightcurves is True:
            least_lightcurve_differences = min(self._difference_values_per_lightcurve)
        else:
            least_lightcurve_differences = self._argument_container.number_lightcurve_samples

        for lc in self._lightcurves:
            lc.select_difference_samples(least_lightcurve_differences, random_generator=random_generator)

    def _get_difference_values_per_lightcurve(self):
        """Retrieves the number of difference values per lightcurve and stores
        them in an array.
        """
        self._difference_values_per_lightcurve = [lc.number_of_difference_values for lc in self._lightcurves]

    def _bin_dts(self, dts):
        """Bin an input array of delta times (dts). Supports several binning
        schemes.

        Parameters
        ----------
        dts : `numpy.ndarray` (N,)
            1-d array of delta times to bin

        Returns
        -------
        bins : `numpy.ndarray` (N,)
            The returned bins array.
        """

        num_bins = int(np.ceil(len(dts) / self._bin_count_target))
        dts_unique = np.unique(dts)
        if self._binning_method == "size":
            quantiles = np.linspace(0.0, 1.0, num_bins + 1)
            self._bins = np.quantile(dts_unique, quantiles)

        elif self._binning_method == "length":
            # Compute num_bins equally spaced bins.
            min_val = dts_unique.min()
            max_val = dts_unique.max()
            self._bins = np.linspace(min_val, max_val, num_bins + 1)

            # Extend the start of the first bin by 0.1% of the range to
            # include the first element. Note this is also done to match
            # Panda's cut function.
            self._bins[0] -= 0.001 * (max_val - min_val)

        elif self._binning_method == "loglength":
            log_vals = np.log(dts_unique)

            # Compute num_bins equally spaced bins in log space.
            min_val = log_vals.min()
            max_val = log_vals.max()
            self._bins = np.linspace(min_val, max_val, num_bins + 1)

            # Extend the start of the first bin by 0.1% of the range to
            # include the first element. Note this is also done to match
            # Panda's cut function.
            self._bins[0] -= 0.001 * (max_val - min_val)
            self._bins = np.exp(self._bins)

        else:
            raise ValueError(f"Method '{self._binning_method}' not recognized")

    def _calculate_binned_statistics(self, sample_values=None, statistic_to_apply="mean"):
        """This method will bin delta_t values stored in `self._dts` using the
        bin edges defined by `self._bins`. Then the corresponding `sample_values`
        in each bin will have a statistic measure applied.

        Parameters
        ----------
        sample_values : `np.ndarray`, optional
            The values that will be used to calculate the `statistic_to_apply`.
            If None or not provided, will use `self._all_d_fluxes` by default.
        statistic_to_apply : str or function, optional
            The statistic to apply to the values in each delta_t bin, by default
            "mean".

        Returns
        -------
        (`List[float]`, `List[float]`)
            A tuple of two lists.
            The first list contains the mean of the delta_t values in each bin.
            The second list contains the result of evaluating the
            statistic measure on the delta_flux values in each delta_t bin.

        Notes
        -----
        1) Largely speaking this is a wrapper over Scipy's `binned_statistic`,
        so any of the statistics supported by that function are valid inputs here.

        2) It is expected that the shapes of `self._dts` and `sample_values` are
        the same. Additionally, any entry at the i_th index of `self._dts` must
        correspond to the same pair of observations as the entry at the i_th
        index of `sample_values`.
        """

        if sample_values is None:
            sample_values = [lc.sample_d_fluxes for lc in self._lightcurves]

        if len(sample_values) != len(self._lightcurves):
            raise AttributeError("Number of lightcurves must equal sample_values.")

        # combining treats all lightcurves as one when calculating the structure function
        if self._argument_container.combine and len(self._lightcurves) > 1:
            all_sample_delta_times = np.hstack([lc.sample_d_times for lc in self._lightcurves])
            all_binning_delta_times = np.hstack([lc._all_d_times for lc in self._lightcurves])
            all_sample_values = np.hstack([s for s in sample_values])

            # binning
            if self._bins is None:
                self._bin_dts(all_binning_delta_times)

            # structure function at specific dt
            # the line below will throw error if the bins are not covering the whole range
            try:
                sfs, _, _ = binned_statistic(
                    all_sample_delta_times, all_sample_values, statistic=statistic_to_apply, bins=self._bins
                )
            except AttributeError:
                raise AttributeError("Length of all_delta_times must equal length of all_sample_value.")

            # return the mean delta_time values for each bin
            bin_means, _, _ = binned_statistic(
                all_sample_delta_times, all_sample_delta_times, statistic="mean", bins=self._bins
            )
            return [bin_means], [sfs]

        # Not combining calculates structure function for each light curve independently
        else:
            # may want to raise warning if len(times) <=1 and combine was set true
            sfs_all = []
            t_all = []
            for lc_idx in range(len(self._lightcurves)):
                # ! double check this to see if this should be len(...) >= 1.
                if len(self._lightcurves[lc_idx]._all_d_times) > 1:
                    # bin the delta_time values, and evaluate the `statistic_to_apply`
                    # for the delta_flux values in each bin.

                    # If the users has not defined bins, calculate them using all the time difference values
                    if self._bins is None:
                        self._bin_dts(self._lightcurves[lc_idx]._all_d_times)

                    try:
                        sfs, _, _ = binned_statistic(
                            self._lightcurves[lc_idx].sample_d_times,
                            sample_values[lc_idx],
                            statistic=statistic_to_apply,
                            bins=self._bins,
                        )
                    except AttributeError:
                        raise AttributeError(
                            "Length of self._lightcurves[lc_idx].sample_d_times array \
                                must equal length of corresponding sample_value array."
                        )

                    sfs_all.append(sfs)

                    # return the mean delta_time values for each bin
                    bin_means, _, _ = binned_statistic(
                        self._lightcurves[lc_idx].sample_d_times,
                        self._lightcurves[lc_idx].sample_d_times,
                        statistic="mean",
                        bins=self._bins,
                    )

                    t_all.append(bin_means)
                else:
                    sfs_all.append(np.array([]))
                    t_all.append(np.array([]))
            return t_all, sfs_all

    @staticmethod
    @abstractmethod
    def name_id() -> str:
        """This method will return the unique name of the Structure Function
        calculation method.
        """
        raise (NotImplementedError, "Must be implemented as a static method by the child class")

    @staticmethod
    @abstractmethod
    def expected_argument_container() -> type:
        """This method will return the argument container class type (not an
        instance) that the Structure Function calculation method requires in
        order to perform it's calculations.
        """
        raise (NotImplementedError, "Must be implemented as a static method by the child class")
